# Copyright (c) 2016-2022 Association of Universities for Research in Astronomy, Inc. (AURA)
# For license information see LICENSE or https://opensource.org/licenses/BSD-3-Clause

from dataclasses import dataclass
from datetime import timedelta
from enum import Enum


class TimeAccountingCode(str, Enum):
    """
    The time accounting codes for the possible partner submissions or internal program
    types used at Gemini, also known as categories.

    This will have to be customized for a given observatory if used independently
    of Gemini.
    """
    AR = 'Argentina'
    AU = 'Australia'
    BR = 'Brazil'
    CA = 'Canada'
    CFH = 'CFHT Exchange'
    CL = 'Chile'
    KR = 'Republic of Korea'
    DD = "Director's Time"
    DS = 'Demo Science'
    GS = 'Gemini Staff'
    GT = 'Guaranteed Time'
    JP = 'Subaru'
    LP = 'Large Program'
    LTP = 'Limited-term Participant'
    SV = 'System Verification'
    UH = 'University of Hawaii'
    UK = 'United Kingdom'
    US = 'United States'
    XCHK = 'Keck Exchange'


@dataclass
class TimeAllocation:
    """
    Time allocation information for a given category for a program.
    Programs may be sponsored by multiple categories with different amounts
    of time awarded. This class maintains information about the time awarded
    and the time that has been used, divided between program time and partner
    calibration time. The time used is calculated as a ratio of the awarded time
    for this category to the total time awarded to the program.

    Attribute:
        category (TimeAccountingCode):
        program_awarded (timedelta):
        partner_awarded (timedelta):
        program_used (timedelta):
        partner_used (timedelta):
    """
    category: TimeAccountingCode
    program_awarded: timedelta
    partner_awarded: timedelta
    program_used: timedelta
    partner_used: timedelta

    def total_awarded(self) -> timedelta:
        return self.program_awarded + self.partner_awarded

    def total_used(self) -> timedelta:
        return self.program_used + self.partner_used

    def __hash__(self):
        return self.category.__hash__()



# Copyright (c) 2016-2022 Association of Universities for Research in Astronomy, Inc. (AURA)
# For license information see LICENSE or https://opensource.org/licenses/BSD-3-Clause

from abc import ABC
from dataclasses import dataclass


@dataclass(frozen=True)
class ID(ABC):
    id: str

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(id={self.id})'

    def __le__(self, other) -> bool:
        return self.id < other.id

    def __lt__(self, other) -> bool:
        return self.id <= other.id

    def __ge__(self, other) -> bool:
        return self.id > other.id

    def __gt__(self, other) -> bool:
        return self.id >= other.id

    def __eq__(self, other) -> bool:
        return type(self) == type(other) and self.id == other.id


@dataclass(frozen=True)
class ProgramID(ID):
    ...


@dataclass(frozen=True)
class GroupID(ID):
    ...


@dataclass(frozen=True)
class UniqueGroupID(ID):
    ...


@dataclass(frozen=True)
class ObservationID(ID):
    @property
    def to_unique_group_id(self) -> UniqueGroupID:
        return UniqueGroupID(self.id)

    def program_id(self) -> ProgramID:
        """Return program ID string from observation ID string"""
        return ProgramID(self.id[0:self.id.rfind('-')])



# Copyright (c) 2016-2022 Association of Universities for Research in Astronomy, Inc. (AURA)
# For license information see LICENSE or https://opensource.org/licenses/BSD-3-Clause

from datetime import timedelta
from typing import Final

from lucupy.minimodel import Observation

from ...types import ZeroTime
from .geminiproperties import GeminiProperties

_IGRINS_CAL_TIME: Final[timedelta] = timedelta(minutes=10)


def with_igrins_cal(func):
    def add_calibration(self):
        if (GeminiProperties.Instruments.IGRINS in self.required_resources()
                and self.total_used - self.program_time() > ZeroTime):
            return func(self) + _IGRINS_CAL_TIME
        return func(self)

    return add_calibration


class GeminiObservation(Observation):
    """A Gemini-specific extension of the Observation class.
    """

    @with_igrins_cal
    def total_used(self) -> timedelta:
        """Override total_used method from Observation.

           Adds IGRINS calibration time using a decorator.

        Returns:
            timedelta: Total used time.
        """
        return super().total_used()



# Copyright (c) 2016-2023 Association of Universities for Research in Astronomy, Inc. (AURA)
# For license information see LICENSE or https://opensource.org/licenses/BSD-3-Clause

from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from datetime import timedelta
from enum import Enum, auto
from typing import Final, FrozenSet, List, Optional

from lucupy.helpers import flatten
from lucupy.minimodel.constraints import Constraints
from lucupy.minimodel.ids import (GroupID, ObservationID, ProgramID,
                                  UniqueGroupID)
from lucupy.minimodel.observation import Observation, ObservationClass
from lucupy.minimodel.resource import Resources
from lucupy.minimodel.site import Site
from lucupy.minimodel.wavelength import Wavelengths

ROOT_GROUP_ID: Final[GroupID] = GroupID('root')


@dataclass
class Group(ABC):
    """This is the base implementation of AND / OR Groups.
    Python does not allow classes to self-reference unless in static contexts,
    so we make a very simple base class to self-reference from subclasses since
    we need this functionality to allow for group nesting.


    Attributes:
        id (GroupID): the identifier of the group.
        program_id (ProgramID): the identifier of the program to which this group belongs.
        group_name (str): a human-readable name of the group.
        number_to_observe (int): the number of children in the group that must be observed for the group to be
                                 considered complete.
        delay_min(timedelta): used in cadences.
        delay_max(timedelta): used in cadences.
        children (Union[List[Group], Observation]): member(s) of the group
    """
    id: GroupID
    program_id: ProgramID
    group_name: str
    number_to_observe: int
    delay_min: timedelta
    delay_max: timedelta
    children: List[Group] | Observation

    # Calculated afterward.
    unique_id: UniqueGroupID = field(init=False)

    def __post_init__(self):
        if self.number_to_observe <= 0:
            msg = f'Group {self.group_name} specifies non-positive {self.number_to_observe} children to be observed.'
            raise ValueError(msg)

        if self.id.id.startswith(self.program_id.id):
            unique_id = self.id.id
        else:
            unique_id = f'{self.program_id.id}:{self.id.id}'
        object.__setattr__(self, 'unique_id', UniqueGroupID(unique_id))

    def subgroup_ids(self) -> FrozenSet[GroupID]:
        """Get the ids for all the subgroups inside. This should probably not be used.

        Returns:
            FrozenSet[GroupID]: Set of GroupID values.
        """

        def aux(group: Group) -> FrozenSet[GroupID]:
            """
            Private method to amalgamate all the UniqueIDs starting at a Group.
            """
            ids = {group.id}
            if not isinstance(group.children, Observation):
                ids |= aux(*[sg for sg in group.children])
            return frozenset(ids)
        return aux(self)

    def subgroup_unique_ids(self) -> FrozenSet[UniqueGroupID]:
        """Get all the unique ids for all the subgroups inside.

        Returns:
            FrozenSet[UniqueGroupID]: Consists of all the UniqueGroupID for Groups rooted here.
        """
        def aux(group: Group) -> FrozenSet[UniqueGroupID]:
            """
            Private method to amalgamate all the UniqueIDs starting at a Group.
            """
            ids = {group.unique_id}
            if not isinstance(group.children, Observation):
                ids |= aux(*[sg for sg in group.children])
            return frozenset(ids)
        return aux(self)

    def sites(self) -> FrozenSet[Site]:
        """All belonging Sites.

        Returns:
            FrozenSet[Site]: Set of Sites for all observations.
        """
        if isinstance(self.children, Observation):
            return frozenset([self.children.site])
        else:
            return frozenset.union(*[s.sites() for s in self.children])

    def required_resources(self) -> Resources:
        """
        Returns:
            Resources: A set of Resources.
        """
        if isinstance(self.children, Observation):
            return self.children.required_resources()
        else:
            return frozenset(r for c in self.children for r in c.required_resources())

    def wavelengths(self) -> Wavelengths:
        """
        Returns:
            Wavelengths: A set of wavelengths.
        """
        if isinstance(self.children, Observation):
            return self.children.wavelengths()
        else:
            return frozenset(w for c in self.children for w in c.wavelengths())

    def constraints(self) -> FrozenSet[Constraints]:
        """
        Returns:
            FrozenSet[Constraints]: All set of Constraints of children in the group.
        """
        if isinstance(self.children, Observation):
            return frozenset([self.children.constraints])
        else:
            return frozenset(cs for c in self.children for cs in c.constraints())

    def observations(self) -> List[Observation]:
        """
        Returns:
            List[Observation]: A list of Observations.
        """
        if isinstance(self.children, Observation):
            return [self.children]
        else:
            return [o for g in self.children for o in g.observations()]

    def is_observation_group(self) -> bool:
        """
        Returns:
            bool: True if the group is just a single Observation, otherwise False.
        """
        return issubclass(type(self.children), Observation)

    def is_scheduling_group(self) -> bool:
        """
        Returns:
            bool: True if the group is a scheduling group, otherwise False.
        """
        return not self.is_observation_group()

    @property
    def to_observation_id(self) -> ObservationID:
        """
        A method that, for observation groups, returns the corresponding ObservationID for the observation contained
        in the group

        If the group is instead a scheduling group, a TypeError will be raised.
        """
        if self.is_observation_group():
            if isinstance(self.children, Observation):
                return self.children.id
            else:
                raise TypeError('Failed is_observation_group method')
        else:
            raise TypeError('Cannot get an ObservationID from a scheduling group.')

    def exec_time(self) -> timedelta:
        """Total execution time across the children of this group.

        Returns:
            exec_time (timedelta): Sum of all the execution times.
        """
        if isinstance(self.children, Observation):
            return self.children.exec_time()
        else:
            return sum((child.exec_time() for child in self.children), timedelta())

    def prog_time(self) -> timedelta:
        """Program time planned across the group.

        Returns:
            prog_time (timedelta): Sum of all prog_time planned times across children of this group.
        """
        if isinstance(self.children, Observation):
            return self.children.prog_time()
        else:
            return sum((child.prog_time() for child in self.children), timedelta())

    def part_time(self) -> timedelta:
        """Partner time planned across the group.

        Returns:
            part_time (timedelta): Sum of all `part_time` (planned times) across the children of this group.
        """
        if isinstance(self.children, Observation):
            return self.children.part_time()
        else:
            return sum((child.part_time() for child in self.children), timedelta())

    def program_used(self) -> timedelta:
        """Program time used across the group.

        Returns:
            program_used (timedelta): Sum of all program_used times across children of this group.
        """
        if isinstance(self.children, Observation):
            return self.children.program_used()
        else:
            return sum((child.program_used() for child in self.children), timedelta())

    def partner_used(self) -> timedelta:
        """Partner time used across the group.

        Returns:
            partner_used (timedelta): Sum of all `partner_used` across the children of this group.
        """
        if isinstance(self.children, Observation):
            return self.children.partner_used()
        else:
            return sum((child.partner_used() for child in self.children), timedelta())

    def total_used(self) -> timedelta:
        """Total time used across the group: includes program time and partner time.

        Returns:
            total_used (timedelta): Sum of total_used times across all children.
        """
        if isinstance(self.children, Observation):
            return self.children.total_used()
        else:
            return sum((child.total_used() for child in self.children), timedelta())

    def program_observations(self) -> List[Observation]:
        """Return the list of program (science + program calibration) observations in the group"""
        observations = []
        for obs in self.observations():
            if obs.obs_class in [ObservationClass.SCIENCE, ObservationClass.PROGCAL]:
                observations.append(obs)
        return observations

    def partner_observations(self) -> List[Observation]:
        """Return the list of partner calibration observations in the group"""
        observations = []
        for obs in self.observations():
            if obs.obs_class == ObservationClass.PARTNERCAL:
                observations.append(obs)
        return observations

    def daycal_observations(self) -> List[Observation]:
        """Return the list of daytime calibration observations in the group"""
        observations = []
        for obs in self.observations():
            if obs.obs_class == ObservationClass.DAYCAL:
                observations.append(obs)
        return observations

    def obs_class(self) -> ObservationClass:
        """
        Determine a group 'obs_class' for observation groups
        """
        obs_class = ObservationClass.NONE
        if self.is_and_group():
            if isinstance(self.children, Observation):
                obs_class = self.children.obs_class
            else:
                obs_class = min([child.obs_class() for child in self.children])
        return obs_class

    def show(self, depth: int = 1) -> None:
        """Print content of the Group.

        Args:
            depth (int, optional): depth of the separator. Defaults to 1.
        """
        def sep(indent: int) -> str:
            return '-----' * indent

        # Is this a subgroup or an observation?
        group_type = 'Scheduling Group' if self.is_scheduling_group() else 'Observation Group'
        print(f'{sep(depth)} Group: {self.id.id}, unique_id={self.unique_id.id} '
              f'({group_type}, num_children={len(self.children)})')
        if isinstance(self.children, Observation):
            self.children.show(depth + 1)
        else:
            for child in self.children:
                child.show(depth + 1)

    @abstractmethod
    def is_and_group(self) -> bool:
        ...

    @abstractmethod
    def is_or_group(self) -> bool:
        ...

    def __len__(self):
        return 1 if self.is_observation_group() else len(self.children)


class AndOption(Enum):
    """
    Different options available for ordering AND group children.
    CUSTOM is used for cadences.

    Members:
       - CONSEC_ORDERED
       - CONSEC_ANYORDER
       - NIGHT_ORDERED
       - NIGHT_ANYORDER
       - ANYORDER
       - CUSTOM
    """
    CONSEC_ORDERED = auto()
    CONSEC_ANYORDER = auto()
    NIGHT_ORDERED = auto()
    NIGHT_ANYORDER = auto()
    ANYORDER = auto()
    CUSTOM = auto()


@dataclass
class AndGroup(Group):
    """The concrete implementation of an AND group.

    Attributes:
        group_option (AndOption): Specify how its observations should be handled.
        previous (int, optional): An index into the group's children to indicate the previously observed child,
            or None if none of the children have yet been observed. Default to None.

    """
    group_option: AndOption
    previous: Optional[int] = None

    def __post_init__(self):
        super().__post_init__()
        if self.number_to_observe != len(self.children):
            msg = f'AND group {self.group_name} specifies {self.number_to_observe} children to be observed but has ' \
                  f'{len(self.children)} children.'
            raise ValueError(msg)
        if self.previous is not None and (self.previous < 0 or self.previous >= len(self.children)):
            msg = f'AND group {self.group_name} has {len(self.children)} children and an illegal previous value of ' \
                  f'{self.previous}'
            raise ValueError(msg)

    def is_and_group(self) -> bool:
        return True

    def is_or_group(self) -> bool:
        return False

    def instruments(self) -> Resources:
        """
        Returns:
            instruments (Resources): A set of all instruments used in this group.
        """
        if isinstance(self.children, Observation):
            instrument = self.children.instrument()
            if instrument is not None:
                return frozenset({instrument})
            else:
                return frozenset()
        else:
            return frozenset(flatten([child.instruments() for child in self.children]))  # type: ignore


@dataclass
class OrGroup(Group):
    """
    The concrete implementation of an OR group.
    The restrictions on an OR group is that it must explicitly require not all
    of its children to be observed.
    """

    def __post_init__(self):
        super().__post_init__()
        if self.number_to_observe >= len(self.children):
            msg = f'OR group {self.group_name} specifies {self.number_to_observe} children to be observed but has ' \
                  f'{len(self.children)} children.'
            raise ValueError(msg)

    def is_and_group(self) -> bool:
        return False

    def is_or_group(self) -> bool:
        return True



# Copyright (c) 2016-2023 Association of Universities for Research in Astronomy, Inc. (AURA)
# For license information see LICENSE or https://opensource.org/licenses/BSD-3-Clause

# This has to be outside of Observation to avoid circular imports.
from enum import Enum
from typing import FrozenSet


class ObservationMode(str, Enum):
    # TODO: This is not stored anywhere and is only used temporarily in the atom code in the
    # TODO: OcsProgramExtractor. Should it be stored anywhere or is it only used in intermediate
    # TODO: calculations? It seems to depend on the instrument and FPU.
    UNKNOWN = 'unknown'
    IMAGING = 'imaging'
    LONGSLIT = 'longslit'
    IFU = 'ifu'
    MOS = 'mos'
    XD = 'xd'
    CORON = 'coron'
    NRM = 'nrm'


ObservationModes = FrozenSet[ObservationMode]



# Copyright (c) 2016-2023 Association of Universities for Research in Astronomy, Inc. (AURA)
# For license information see LICENSE or https://opensource.org/licenses/BSD-3-Clause

from dataclasses import dataclass, field
from datetime import timedelta

from .qastate import QAState
from .resource import Resources
from .wavelength import Wavelengths


@dataclass
class Atom:
    """
    Atom information, where an atom is the smallest schedulable set of steps
    such that useful science can be obtained from performing them.
    Wavelengths must be specified in microns.

    Attributes:

        id (int): GPP atom `id`. In other case is given by the Provider
        exec_time (timedelta): Total planned time of execution.
        prog_time (timedelta): Planned program time.
        part_time (timedelta): Planned partner time.
        program_used (timedelta): Used (charged) program time.
        partner_used (timedelta): Used (charged) partner time.
        observed (bool): True if the STATUS is already observed.
        qa_state (QAState):
        guide_state (bool): True if a state exists.
        resources (Resources): Resources needed (Instrument, FPU, etc).
        wavelengths (Wavelengths): Set of wavelengths.

    """
    id: int
    exec_time: timedelta = field(hash=False, compare=False)
    prog_time: timedelta = field(hash=False, compare=False)
    part_time: timedelta = field(hash=False, compare=False)
    program_used: timedelta = field(hash=False, compare=False)
    partner_used: timedelta = field(hash=False, compare=False)
    observed: bool
    qa_state: QAState
    guide_state: bool
    resources: Resources
    wavelengths: Wavelengths



# Copyright (c) 2016-2023 Association of Universities for Research in Astronomy, Inc. (AURA)
# For license information see LICENSE or https://opensource.org/licenses/BSD-3-Clause

import sys
from datetime import datetime
from typing import Tuple


def sex2dec(stime: str,
            todegree: bool = False,
            sep: str = ':') -> float:
    """"Sexadecimal to decimal

    Args:
        stime (str): A string of format "HR:MIN:SEC
        todegree (bool, optional): Option to convert to degree. Defaults to False.
        sep (str, optional): Separator. Defaults to ':'.

    Raises:
        ValueError: Wrong format for not following the separator convention.

    Returns:
        float: The decimal equivalent
    """
    l_stime = str(stime).replace("+", "")
    if sep not in l_stime:
        raise ValueError(f'Separator {sep} not found in {stime}. Input must be in the format "HR<sep>MIN<sep>SEC"')

    f = 1.0
    if todegree:
        f = 15.0

    result = 0.0
    exp = 0
    sign = 1.
    for val in l_stime.split(sep):
        x = float(val)
        if x < 0.0:
            sign = -1.
        result += abs(x) / 60. ** exp
        exp += 1
    return sign * f * result


def dtsex2dec(dt: datetime,
              todegree: bool = False) -> float:

    """Datetime to decimals

    Returns:
        float: The decimal equivalent
    """
    f = 1.0
    if todegree:
        f = 15.0

    sign = 1.
    if dt.hour < 0:
        sign = -1.
    return sign * f * (abs(dt.hour) + dt.minute / 60. + dt.second / 3600.)


def sixty(dd: float) -> Tuple[float, float, float]:
    """Sixty

    Args:
        dd (float): Days

    Returns:
        Tuple[float, float, float]: Degrees, minutes and seconds
    """
    is_positive = dd >= 0
    l_dd = abs(dd)
    minutes, seconds = divmod(l_dd * 3600, 60)
    degrees, minutes = divmod(minutes, 60)
    if degrees > 0.:
        degrees = degrees if is_positive else -degrees
    elif minutes > 0.:
        minutes = minutes if is_positive else -minutes
    else:
        seconds = seconds if is_positive else -seconds
    return degrees, minutes, seconds


def dec2sex(d: float,
            p: int = 3,
            cutsec: bool = False,
            hour: bool = False,
            tohour: bool = False,
            sep: str = ':',
            leadzero: int = 0,
            round_min: bool = False) -> str:
    """Convert decimal degrees/hours to a formatted sexigesimal string.

    Args:
        d: Input in degrees
        p: Digits for seconds
        cutsec: Cut seconds, just display, e.g. DD:MM
        hour: d is decimal hours, so must be <=24
        tohour: Convert from degress to hours (divide by 15.)
        sep: Separator string
        leadzero: If >0 display leading 0's, e.g. -05:25. The value is the number of digits for the DD or HR field.
        round_min: When cutsec, round to the nearest minute rather than truncate
    Returns:
        str: That modify string
    """
    l_d = float(d)
    sign = ''
    maxdg = 360.

    if tohour:
        l_d /= 15.0
        hour = True

    if hour and (abs(l_d) > 24.):
        print('Input in hours must be less than or equal to 24.0.')
        sys.exit(1)

    if hour:
        maxdg = 24.0

    n = 2 if p == 0 else 3
    secstr = '{:0' + '{:1d}'.format(n + p) + '.' + '{:1d}'.format(p) + 'f}'

    six = sixty(l_d)

    dg = abs(int(six[0]))
    if six[0] < 0:
        sign = '-'

    if leadzero > 0:
        sldg = '0' + str(leadzero)
    else:
        sldg = str(len(str(dg)))

    mn = abs(int(six[1]))
    if six[1] < 0:
        sign = '-'

    sc = float(secstr.format(abs(six[2])))
    if six[2] < 0.0:
        sign = '-'

    if sc >= 60.0:
        sc -= 60.0
        mn += 1

    if cutsec:
        if round_min and sc >= 30.:
            # Round to the nearest minute, otherwise truncate
            mn += 1
        sc = 0.0

    if mn >= 60:
        mn -= 60
        dg += 1

    if dg >= int(maxdg):
        dg -= int(maxdg)

    if cutsec and sc == 0.0:
        fmt = '{:1s}{:' + sldg + 'd}' + sep + '{:02d}'
        s = fmt.format(sign, dg, mn)
    else:
        fmt = '{:1s}{:' + sldg + 'd}' + sep + '{:02d}' + sep + secstr
        s = fmt.format(sign, dg, mn, sc)

    return s.strip()



# All code in this package is a refactored, numpy-vectorized version of thorskyutil.py:
#
# https://github.com/jrthorstensen/thorsky/blob/master/thorskyutil.py
#
# utility and miscellaneous time and the sky routines built mostly on astropy.
#
# Copyright John Thorstensen, 2018, who graciously has allowed Gemini to use this code under the BSD-3 Clause license.
# For license information see LICENSE or https://opensource.org/licenses/BSD-3-Clause

from datetime import datetime, timedelta
from typing import Tuple

import astropy.units as u
import numpy as np
import numpy.typing as npt
from astropy.coordinates import (Angle, BaseRADecFrame, EarthLocation,
                                 PrecessedGeocentric)
from astropy.time import Time
from astropy.units import Quantity
from pytz import BaseTzInfo

from .altitude import AngleParam
from .constants import EQUAT_RAD, FLATTEN, J2000, JYEAR, JYEAR_100


def current_geocent_frame(time: Time) -> BaseRADecFrame:
    """Get current frame for the equinox specified by the time.

    Args:
        time: If an array then the first entry is used

    Returns:
        an astropy PrecessedGeocentric time.
    """
    # Generate a PrecessedGeocentric frame for the current equinox.
    time_ep = 2000. + (np.asarray(time.jd) - J2000) / JYEAR
    if time_ep.ndim == 0:
        time_ep = time_ep[None]  # Makes 1D
    equinox = Time(f'J{time_ep[0]:7.2f}')

    return PrecessedGeocentric(equinox=equinox)


def geocentric_coors(geolong: Angle, geolat: float, height: float) -> Tuple[float, float, float]:
    """Geocentric XYZ coordinates for a location at a longitude, latitude and height above sea level.

    Retained because if one replaces the longitude input with the
    sidereal time, the return is the XYZ in the equatorial frame
    of date.  This is used in the lunar topocentric correction.

    Args:
        geolong: Geographic longitude, or LST to get celestial-aligned result
        geolat:  Geographic latitude
        height: Height above sea level, which must be in meters.

    Returns:
        Triple of distances.
    """

    # computes the geocentric coordinates from the geodetic
    # (standard map-type) longitude, latitude, and height.
    # Notation generally follows 1992 Astr Almanac, p. K11 */
    # NOTE that if you replace "geolong" with the local sidereal
    # time, this automatically gives you XYZ in the equatorial frame
    # of date.
    # In this version, geolong and geolat are assumed to be
    # Angles and height is assumed to be in meters; returns
    # a triplet of explicit Distances.

    denom = (1. - FLATTEN) * np.sin(geolat)
    denom = np.cos(geolat) * np.cos(geolat) + denom * denom
    c_geo = 1. / np.sqrt(denom)
    s_geo = (1. - FLATTEN) * (1. - FLATTEN) * c_geo
    c_geo = c_geo + height / EQUAT_RAD

    #  deviation from almanac notation -- include height here.
    s_geo = s_geo + height / EQUAT_RAD

    # distancemultiplier = Distance(_Constants.EQUAT_RAD, unit = u.m)
    x_geo = EQUAT_RAD * c_geo * np.cos(geolat) * np.cos(geolong)
    y_geo = EQUAT_RAD * c_geo * np.cos(geolat) * np.sin(geolong)
    z_geo = EQUAT_RAD * s_geo * np.sin(geolat)

    return x_geo, y_geo, z_geo


def min_max_alt(lat: Angle, dec: AngleParam) -> Tuple[Angle, Angle]:
    """Finds the minimum and maximum altitudes of a celestial location.

    Args:
        lat (Angle) : Latitude of site.
        dec (AngleParam) : Declination of the object.

    Returns:
       minalt, maxalt (Tuple[Angle, Angle]): tuple of minimum and maximum altitudes.
    """

    # arguments are Angles; returns (minalt, maxalt) as Angles
    # where min and max are the minimum and maximum altitude
    # an object at declination dec reaches at this latitude.

    dec = np.asarray(dec.to_value(u.rad).data)  # type: ignore
    scalar_input = False
    if dec.ndim == 0:
        dec = dec[None]
        scalar_input = True

    maxalt = np.zeros(len(dec))
    minalt = np.zeros(len(dec))

    x = np.cos(dec) * np.cos(lat) + np.sin(dec) * np.sin(lat)
    ii = np.where(abs(x) <= 1.)[0][:]
    if len(ii) != 0:
        maxalt[ii] = np.arcsin(x[ii])

    x = np.sin(dec) * np.sin(lat) - np.cos(dec) * np.cos(lat)
    ii = np.where(abs(x) <= 1.)[0][:]
    if len(ii) != 0:
        minalt[ii] = np.arcsin(x[ii])

    if scalar_input:
        minalt = np.squeeze(minalt)
        maxalt = np.squeeze(maxalt)
    return Angle(minalt, unit=u.rad), Angle(maxalt, unit=u.rad)


def local_midnight_time(time: Time, localtzone: BaseTzInfo) -> Time:
    """Find nearest local midnight (UT).

    If it's before noon local time, returns previous midnight;
    if after noon, return next midnight.

    Args:
        time (Time): astropy Time object
        localtzone (timezone) : Timezone object.

    Returns:
        Time. This is not zone-aware, but should be correct.
    """

    # takes an astropy Time and the local time zone and
    # generates another Time which is the nearest local
    # clock-time midnight.  The returned Time is unaware
    # of the timezone but should be correct.

    time = Time(np.asarray(time.iso), format='iso')
    scalar_input = False
    if time.ndim == 0:
        time = time[None]  # Makes 1D
        scalar_input = True

    datetmid = []
    for time in time:
        datet = time.to_datetime(timezone=localtzone)

        # if before midnight, add 12 hours
        if datet.hour >= 12:
            datet = datet + timedelta(hours=12.)
        datetmid.append(localtzone.localize(datetime(datet.year, datet.month, datet.day, 0, 0, 0)))

    if scalar_input:
        result = Time(datetmid[0])
    else:
        result = Time(datetmid)
    return result


def local_sidereal_time(time: Time, location: EarthLocation) -> Angle:
    """Moderate-precision (1 sec) local sidereal time.

    Adapted with minimal changes from skycalc routine. Native
    astropy routines are unnecessarily precise for our purposes and
    rather slow.

    Args:
        time (Time): Time at which to compute the local sidereal time.
        location (Location): Location on Earth for which to compute the local sidereal time.

    Returns:
        lst (Angle): Local sidereal time.
    """
    # julian date represented as integer values
    julian_int = np.asarray(time.jd, dtype=int)

    scalar_input = False
    # check if time is an array or a scalar
    if julian_int.ndim == 0:
        julian_int = julian_int[None]
        scalar_input = True

    fraction = time.jd - julian_int
    mid = julian_int + 0.5
    ut = fraction - 0.5
    less_than_half = np.where(fraction < 0.5)[0][:]
    if len(less_than_half) != 0:
        mid[less_than_half] = julian_int[less_than_half] - 0.5
        ut[less_than_half] = fraction[less_than_half] + 0.5  # as fraction of a day.

    t = (mid - J2000) / JYEAR_100

    sidereal = (24110.54841 + 8640184.812866 * t + 0.093104 * t ** 2 - 6.2e-6 * t ** 3) / 86400.
    # at Greenwich
    sid_int = sidereal.astype(np.int64)
    sidereal = sidereal - sid_int
    # longitude is measured east so add.
    sidereal = sidereal + 1.0027379093 * ut + location.lon.hour / 24.

    sid_int = sidereal.astype(np.int64)
    sidereal = (sidereal - sid_int) * 24.

    # TODO: A conversion to radians is needed if the output is not an Angle
    lst = Angle(sidereal, unit=u.hour)
    lst.wrap_at(24. * u.hour, inplace=True)

    if scalar_input:
        return lst.squeeze()
    return lst


def true_airmass(altit: Angle) -> npt.NDArray[float]:
    """True airmass for an altitude.
    Equivalent of getAirmass in the QPT, based on vskyutil.true_airmass
    https://github.com/gemini-hlsw/ocs/blob/12a0999bc8bb598220ddbccbdbab5aa1e601ebdd/bundle/edu.gemini.qpt.client/src/main/java/edu/gemini/qpt/core/util/ImprovedSkyCalcMethods.java#L119

    Based on a fit to Kitt Peak airmass tables, C. M. Snell & A. M. Heiser, 1968,
    PASP, 80, 336.  Valid to about airmass 12, and beyond that just returns
    secz minus 1.5, which won't be quite right.

    Takes an Angle and return the true airmass, based on a tabulation of the mean KPNO
    atmosphere given by C. M. Snell & A. M. Heiser, 1968, PASP, 80, 336.
    They tabulated the airmass at 5 degr intervals from z = 60 to 85 degrees; The data was fit  with
    a fourth order poly for (secz - airmass) as a function of (secz - 1) using
    the IRAF curfit routine, then adjusted the zeroth order term to force (secz - airmass) to zero at
    z = 0.  The poly fit is very close to the tabulated points (largest difference is 3.2e-4)
    and appears smooth. This 85-degree point is at secz = 11.47, so for secz > 12 just return secz

    coefs = [2.879465E-3,  3.033104E-3, 1.351167E-3, -4.716679E-5]

    Args:
        altit: Altitude above horizon.

    Returns:
        True airmass value
    """
    altit = np.asarray(altit.to_value(u.rad).data)
    scalar_input = False
    if altit.ndim == 0:
        altit = altit[None]  # Makes 1D
        scalar_input = True

    # ret = np.zeros(len(altit))
    ret = np.full(len(altit), 500.)
    ii = np.where(altit > 0.0)[0][:]
    if len(ii) != 0:
        ret[ii] = 1. / np.sin(altit[ii])  # sec z = 1/sin (altit)

    kk = np.where(np.logical_and(ret >= 0.0, ret < 12.))[0][:]
    if len(kk) != 0:
        seczmin1 = ret[kk] - 1.
        coefs = np.array([-4.716679E-5, 1.351167E-3, 3.033104E-3, 2.879465E-3, 0.])
        ret[kk] -= np.polyval(coefs, seczmin1)

    if scalar_input:
        return np.squeeze(ret)
    return ret


def hour_angle_to_angle(dec: AngleParam,
                        lat: AngleParam,
                        alt: AngleParam) -> Angle:
    """Transform the Hour Angle in to an Angle.

    The hour angle from spherical astronomy, eastor west of meridian,
    not the u.hourangle from astropy.

    dec and alt must have the same dimensions.

    If the object is always above alt, an Angle of +1000 radians is returned.
    If always below, -1000 radians.

    Args:
        dec: Declination of source.
        lat: Latitude of site.
        alt: Height above horizon for computation.

    Returns:
        An Angle giving the hour angle at which the declination dec reaches altitude alt.

    """

    # Arguments are all angles.
    # returns hour angle at which object at dec is at altitude alt for a
    # latitude lat.

    dec = np.asarray(dec.to_value(u.rad).data) * u.rad  # type: ignore
    alt = np.asarray(alt.to_value(u.rad)) * u.rad  # type: ignore

    scalar_input = False
    if dec.ndim == 0 and alt.ndim == 0:
        scalar_input = True
    if dec.ndim == 0:
        dec = dec[None]
    if alt.ndim == 0:
        alt = alt[None]

    if len(dec) == 1 and len(alt) > 1:
        dec *= np.ones(len(alt))
    elif len(dec) > 1 and len(alt) == 1:
        alt *= np.ones(len(dec))
    elif len(dec) != len(alt):
        raise ValueError('Error: dec and alt have incompatible lengths')

    x = np.zeros(len(dec))
    codec = np.zeros(len(dec))
    zdist = np.zeros(len(dec))

    minalt, maxalt = min_max_alt(lat, dec)
    ii = np.where(alt < minalt)[0][:]
    if len(ii) != 0:
        x[ii] = -1000.

    jj = np.where(alt > maxalt)[0][:]
    if len(jj) != 0:
        x[jj] = 1000.

    kk = np.where(np.logical_and(alt >= minalt, alt <= maxalt))[0][:]
    if len(kk) != 0:
        rightang = Angle(np.pi / 2, unit=u.rad)
        codec[kk] = rightang - dec[kk]
        colat = rightang - lat
        zdist[kk] = rightang - alt[kk]
        x[kk] = (np.cos(zdist[kk]) - np.cos(codec[kk]) * np.cos(colat)) / (np.sin(codec[kk]) * np.sin(colat))
        x[kk] = np.arccos(x[kk])

    if scalar_input:
        # return (Angle(np.squeeze(x), unit = u.rad))
        x = np.squeeze(x)
    return Angle(x, unit=u.rad)


def xair(zd: Quantity) -> npt.NDArray[float]:
    """Evaluate true airmass.

    Equation 3 from Krisciunas &  Schaefer 1991.
    Trick for handling arrays and scalars from
    https://stackoverflow.com/questions/29318459/python-function-that-handles-scalar-or-arrays

    Args:
        zd: Float or numpy array of zenith distance angles

    Returns:
        npt.NDArray[float]: values converted.

    """

    zd = np.asarray(zd.to_value(u.rad).data) * u.rad
    scalar_input = False
    if zd.ndim == 0:
        zd = zd[None]  # Makes 1D
        scalar_input = True

    x = np.zeros(len(zd))

    # below the horizon
    kk = np.where(zd > 90. * u.deg)[0][:]
    if len(kk) != 0:
        x[kk] = 10.

    # above the horizon
    ik = np.where(zd <= 90. * u.deg)[0][:]
    if len(ik) != 0:
        v = np.sqrt(1.0 - 0.96 * np.sin(zd[ik]) ** 2)
        ii = np.where(v != 0.0)[0][:]
        if len(ii) != 0:
            x[ik[ii]] = 1. / v[ii]
        jj = np.where(v == 0.0)[0][:]
        if len(jj) != 0:
            x[ik[jj]] = 10.

    if scalar_input:
        return x.squeeze()
    return x


def ztwilight(alt: Angle) -> npt.NDArray[float]:
    """Estimate twilight contribution to zenith sky brightness, in magnitudes per square arcsecond.

    Evaluates a polynomial approximation to observational data (see source for
    reference) of zenith sky brightness (blue) as a function of the sun's elevation
    from -0.9 degrees to -18 degrees.  For reference, 3 mag is roughly Nautical
    twilight and looks 'pretty dark'; something like 10 mag is about the maximum
    for broadband sky flats in many cases.

    Args:
        alt (Angle): Sun's elevation.  Meaningful range is -0.9 to -18 degrees.

    Returns:
       An array but if the sun is up, returns 20; if the sun below -18, returns 0.

    """
    # Given an Angle alt in the range -0.9 to -18 degrees,
    # evaluates a polynomial expansion for the approximate brightening
    # in magnitudes of the zenith in twilight compared to its
    # value at full night, as function of altitude of the sun (in degrees).
    # To get this expression I looked in Meinel, A.,
    # & Meinel, M., "Sunsets, Twilight, & Evening Skies", Cambridge U.
    # Press, 1983; there's a graph on p. 38 showing the decline of
    # zenith twilight.  I read points off this graph and fit them with a
    # polynomial.
    # Comparison with Ashburn, E. V. 1952, JGR, v.57, p.85 shows that this
    # is a good fit to his B-band measurements.

    alt = Angle(np.asarray(alt.deg), unit=u.deg)
    scalar_input = False
    if alt.ndim == 0:
        alt = alt[None]  # Makes 1D
        scalar_input = True

    twisb = np.zeros(len(alt))
    # flag for sun up, also not grossly wrong.
    ii = np.where(alt.deg > -0.9)[0][:]
    if len(ii) != 0:
        twisb[ii] = 20.

    # fully dark, no contrib to zenith skyglow.
    jj = np.where(alt.deg < -18.)[0][:]
    if len(jj) != 0:
        twisb[jj] = 0.

    kk = np.where(np.logical_and(alt.deg >= -18., alt.deg <= -0.9))[0][:]
    if len(kk) != 0:
        y = (-1. * alt[kk].deg - 9.0) / 9.0  # my polynomial's argument
        twisb[kk] = ((2.0635175 * y + 1.246602) * y - 9.4084495) * y + 6.132725

    if scalar_input:
        twisb = np.squeeze(twisb)
    return twisb



# All code in this package is a refactored, numpy-vectorized version of thorskyutil.py:
#
# https://github.com/jrthorstensen/thorsky/blob/master/thorskyutil.py
#
# utility and miscellaneous time and the sky routines built mostly on astropy.
#
# Copyright John Thorstensen, 2018, who graciously has allowed Gemini to use this code under the BSD-3 Clause license.
# For license information see LICENSE or https://opensource.org/licenses/BSD-3-Clause

from .altitude import *
from .brightness import *
from .constants import *
from .events import *
from .moon import *
from .sun import *
from .utils import *



# Copyright (c) 2016-2023 Association of Universities for Research in Astronomy, Inc. (AURA)
# For license information see LICENSE or https://opensource.org/licenses/BSD-3-Clause

# Basic type aliases for usefulness.
from datetime import timedelta
from typing import Final, List, TypeVar, Union

import numpy.typing as npt
from astropy.time import Time

T = TypeVar('T')

ScalarOrNDArray = Union[T, npt.NDArray[T]]
TimeScalarOrNDArray = Time | npt.NDArray[float]
ListOrNDArray = List[T] | npt.NDArray[T]

ZeroTime: Final[timedelta] = timedelta()

# Convenient type alias for Interval
Interval = npt.NDArray[int]



# All code in this package is a refactored, numpy-vectorized version of thorskyutil.py:
#
# https://github.com/jrthorstensen/thorsky/blob/master/thorskyutil.py
#
# utility and miscellaneous time and the sky routines built mostly on astropy.
#
# Copyright John Thorstensen, 2018, who graciously has allowed Gemini to use this code under the BSD-3 Clause license.
# For license information see LICENSE or https://opensource.org/licenses/BSD-3-Clause

import numpy as np
import numpy.typing as npt
from astropy import units as u
from astropy.coordinates import Angle, Distance
from astropy.units import Quantity

from lucupy.minimodel import SkyBackground

from .constants import EQUAT_RAD, KZEN
from .utils import xair, ztwilight


def calculate_sky_brightness(moon_phase_angle: Angle,
                             target_moon_angdist: Distance,
                             earth_moon_dist: Distance,
                             moon_zenith_distang: Angle,
                             target_zenith_distang: Angle,
                             sun_zenith_distang: Angle,
                             verbose: bool = False) -> npt.NDArray[float]:

    """Calculate sky brightness based on formulas from Krisciunas & Schaefer 1991.
    Uses array processing

    Args:
        moon_phase_angle: Moon phase angles in degrees
        target_moon_angdist: Angular distances between target and moon
        earth_moon_dist: Distances from the Earth to the Moon
        moon_zenith_distang: Moon zenith distance angles
        target_zenith_distang: Target zenith distance angles
        sun_zenith_distang: Sun zenith distance angles
        verbose: Verbose output flag

    Returns:
        array of float:  Numpy array of sky background magnitudes at target location

    """

    # Constants
    a = 2.51189
    q = 27.78151

    sun_alt = 90.0 * u.deg - sun_zenith_distang  # sun altitude
    if verbose:
        print(f'sun_alt: {sun_alt}')

    # Relative distance between the Earth and the Moon
    norm_moondist = earth_moon_dist / (60.27 * EQUAT_RAD)

    # Number of positions
    n = len(target_zenith_distang)

    # Dark sky zenith V surface brightness
    v_zen = np.ones(n) * 21.587

    # Correction for twilight
    i = np.where(sun_alt > -18.5 * u.deg)[0][:]
    if len(i) != 0:
        v_zen[i] = v_zen[i] - ztwilight(sun_alt[i])

    # zenith sky brightness
    b_zen = 0.263 * a ** (q - v_zen)

    # Sky brightness with no moon at target, scattering due to airmass
    b_sky = b_zen * xair(target_zenith_distang) * 10.0 ** (
            -0.4 * KZEN * (xair(target_zenith_distang) - 1.0))

    # Lunar sky brightness
    b_moon = np.zeros(n)
    istar = np.zeros(n)
    frho = np.zeros(n)

    # Do calculations when the Moon is above or near the horizon
    im = np.where(moon_zenith_distang < 90.8 * u.deg)[0][:]

    # Illuminance with phase angle, eq. 8
    # Correction for lunar distance.
    istar[im] = 10.0 ** (
            -0.4 * (3.84 + 0.026 * abs(moon_phase_angle[im].value) + 4.e-9 * moon_phase_angle[im].value ** 4.)
    ) / norm_moondist[im] ** 2

    # Rough correction for opposition effect
    # 35 per cent brighter at full, effect tapering linearly to
    #   zero at 7 degrees away from full. mentioned peripherally in
    #   Krisciunas and Scheafer, p. 1035. */
    hh = im[np.where(abs(moon_phase_angle[im]) < 7. * u.deg)[0][:]]
    if len(hh):
        istar[hh] *= (1.35 - 0.05 * abs(moon_phase_angle[hh].value))

    # Build eq. 21 piecewise, mAng is rho in skycalc and the paper
    frho[im] = 229087. * (1.06 + np.cos(target_moon_angdist[im]) ** 2.0)

    # Separation > 10 deg
    ii = im[np.where(abs(target_moon_angdist[im]) > 10.0 * u.deg)[0][:]]
    if len(ii):
        frho[ii] += 10.0 ** (6.15 - target_moon_angdist[ii].value / 40.0)

    # Separation between 0.25 and 10 deg
    jj = im[np.where(np.logical_and(abs(target_moon_angdist[im]) > 0.25 * u.deg,
                                    abs(target_moon_angdist[im]) <= 10.0 * u.deg))[0][:]]
    if len(jj):
        frho[jj] += 6.2e7 / (target_moon_angdist[jj].value ** 2.0)  # eq 19

    # If on lunar disk
    kk = im[np.where(abs(target_moon_angdist[im]) <= 0.25 * u.deg)[0][:]]
    if len(kk) != 0:
        frho[kk] += 9.9e8

    # Sky brightness from the Moon in nanoLamberts (B)
    b_moon[im] = frho[im] * istar[im] * 10 ** (-0.4 * KZEN * xair(moon_zenith_distang[im])) * (
            1.0 - 10 ** (-0.4 * KZEN * xair(target_zenith_distang[im]))
    )

    # hh = np.where(np.logical_and(cc > 0.5, cc < 0.8))[0][:]
    # if len(hh) != 0:  # very simple increase in SB if there are thin clouds
    #     b_moon[hh] = 2.0 * b_moon[hh]

    # sky brightness in Vmag/arcsec^2
    skybright = q - np.log10((b_moon + b_sky) / 0.263) / np.log10(a)

    if verbose:
        print(f'v_zen: {v_zen}')
        print(f'b_zen: {b_zen}')
        print(f'istar: {istar}')
        print(f'b_moon: {b_moon}')
        print(f'b_sky: {b_sky}')
        print(f'skybright: {skybright}')

    return skybright


def calculate_sky_brightness_qpt(moon_phase_angle: Quantity,
                                 target_moon_angdist: Quantity,
                                 moon_zenith_distang: Quantity,
                                 target_zenith_distang: Quantity,
                                 sun_zenith_distang: Quantity,
                                 verbose: bool = False) -> npt.NDArray[float]:
    """
    Calculate sky brightness based on formulas from Krisciunas & Schaefer 1991

    Args:
        moon_phase_angle: Moon phase angle at solar midnight in degrees
        target_moon_angdist: Numpy array of angular distances between target and moon
        moon_zenith_distang: Numpy array of Moon zenith distance angles
        target_zenith_distang: Numpy array of target zenith distance angles
        sun_zenith_distang: Numpy array of Sun zenith distance angles
        verbose: Verbose output flag

    Returns:
        float: Numpy array of sky background magnitudes at target location

    """

    # Constants
    a = 2.51189
    q = 27.78151

    mpaa = np.asarray(moon_phase_angle.value)
    if mpaa.ndim == 0:
        mpaa = np.ones(len(moon_zenith_distang)) * moon_phase_angle.value

    sun_alt = 90.0 * u.deg - sun_zenith_distang  # sun altitude
    if verbose:
        print(f'sun_alt: {sun_alt}')

    # Dark sky zenith V surface brightness
    v_zen = np.ones(len(target_zenith_distang)) * 21.587
    # correction for twilight
    ii = np.where(sun_alt > -18.5 * u.deg)[0][:]
    v_zen[ii] = v_zen[ii] - ztwilight(sun_alt[ii])

    # zenith sky brightness
    b_zen = 0.263 * a ** (q - v_zen)

    # Sky brightness with no moon at target, scattering due to airmass
    b_sky = b_zen * xair(target_zenith_distang) * 10.0 ** (
            -0.4 * KZEN * (xair(target_zenith_distang) - 1.0)
    )

    # Lunar sky brightness
    n = len(b_sky)
    b_moon = np.zeros(n)

    istar = 10.0 ** (-0.4 * (3.84 + 0.026 * abs(mpaa) + 4.e-9 * mpaa ** 4.))

    ii = np.where(moon_zenith_distang < 90.8 * u.deg)[0][:]

    jj = ii[np.where(target_moon_angdist[ii] >= 10.0 * u.deg)[0][:]]
    if len(jj):
        fpjj = (1.06 + np.cos(target_moon_angdist[jj]) ** 2) * 10.0 ** 5.36 + 10.0 ** (
                6.15 - target_moon_angdist[jj].value / 40.0
        )
        b_moon[jj] = fpjj * istar[jj] * 10 ** (-0.4 * KZEN * xair(moon_zenith_distang[jj])) * (
                1.0 - 10 ** (-0.4 * KZEN * xair(target_zenith_distang[jj]))
        )

    kk = np.where(ii != jj)[0][:]
    if len(kk):
        # There is a bug in the following line from the original code, used by QPT
        fpkk = 6.2e7 / (target_moon_angdist[kk].value ** 2)
        # fpkk = (1.06 + np.cos(mdist[kk])**2) * 10.0**5.36 + 6.2e7 / (mdist[kk].value**2)
        b_moon[kk] = fpkk * istar[kk] * 10 ** (-0.4 * KZEN * xair(moon_zenith_distang[kk])) * (
                1.0 - 10 ** (-0.4 * KZEN * xair(target_zenith_distang[kk]))
        )

    # hh = np.where(np.logical_and(cc > 0.5, cc < 0.8))[0][:]
    # if len(hh) != 0:  # very simple increase in SB if there are thin clouds
    #     b_moon[hh] = 2.0 * b_moon[hh]

    skybright = q - np.log10((b_moon + b_sky) / 0.263) / np.log10(a)  # sky brightness in Vmag/arcsec^2

    if verbose:
        print(f'v_zen: {v_zen}')
        print(f'b_zen: {b_zen}')
        print(f'istar: {istar}')
        print(f'b_moon: {b_moon}')
        print(f'b_sky: {b_sky}')
        print(f'skybright: {skybright}')

    return skybright


def convert_to_sky_background(sb: npt.NDArray[float]) -> npt.NDArray[SkyBackground]:
    """Convert visible sky background magnitudes to decimal conditions.

    Args:
        sb:  TargetInfo object with time dependent vsb quantities

    Return:
        npt.NDArray[SkyBackground]: Sky background condition values

    Example:
        Conversion scheme:
            1.0 |          vsb <= 19.61
            0.8 | 19.61 < vsb <= 20.78
            0.5 | 20.78 < vsb <= 21.37
            0.2 | 21.37 < vsb

    """

    cond = np.full([len(sb)], SkyBackground.SBANY)

    ii = np.where(np.logical_and(sb > 19.61, sb <= 20.78))[0][:]
    cond[ii] = SkyBackground.SB80

    ii = np.where(np.logical_and(sb > 20.78, sb <= 21.37))[0][:]
    cond[ii] = SkyBackground.SB50

    ii = np.where(sb > 21.37)[0][:]
    cond[ii] = SkyBackground.SB20

    return cond



# Copyright (c) 2016-2022 Association of Universities for Research in Astronomy, Inc. (AURA)
# For license information see LICENSE or https://opensource.org/licenses/BSD-3-Clause

from .geminiobservation import *
from .geminiproperties import *



# Copyright (c) 2016-2022 Association of Universities for Research in Astronomy, Inc. (AURA)
# For license information see LICENSE or https://opensource.org/licenses/BSD-3-Clause

from enum import IntEnum, auto


class TooType(IntEnum):
    """
    The target-of-opportunity type for a program and for an observation.
    These are ordered specifically so that we can compare them.

    The INTERRUPT is considered the highest level of TooType, followed by RAPID, and then STANDARD.

    Thus, a Program with a RAPID type, for example, can contain RAPID and STANDARD Observations,
    but not INTERRUPT ones.

    The values and ordering on them should NOT be changed as this will break functionality.

    Members:
        - STANDARD
        - RAPID
        - INTERRUPT

    """
    STANDARD = auto()
    RAPID = auto()
    INTERRUPT = auto()



# Copyright (c) 2016-2023 Association of Universities for Research in Astronomy, Inc. (AURA)
# For license information see LICENSE or https://opensource.org/licenses/BSD-3-Clause

from dataclasses import dataclass
from typing import FrozenSet, Optional

from ..decorators import immutable


@immutable
@dataclass(frozen=True)
class Resource:
    """This is a general observatory resource.
    It can consist of a guider, an instrument, or a part of an instrument,
    or even a personnel and is used to determine what observations can be
    performed at a given time based on the resource availability.

    Attributes:
        id (str): Resource id or name.
        description (str): Short description.
    """
    id: str
    description: Optional[str] = None

    def __post_init__(self):
        if self.id is None or 'NONE' in self.id.upper():
            raise ValueError('Should not have any Resources equal to None or containing "None"')

    def __eq__(self, other):
        return isinstance(other, Resource) and self.id == other.id

    def __repr__(self):
        return f"Resource(id='{self.id}')"


Resources = FrozenSet[Resource]

NIR_INSTRUMENTS: Resources = frozenset([Resource('Flamingos2'),
                                        Resource('GNIRS'),
                                        Resource('NIRI'),
                                        Resource('NIFS'),
                                        Resource('IGRINS')])



# Copyright (c) 2016-2023 Association of Universities for Research in Astronomy, Inc. (AURA)
# For license information see LICENSE or https://opensource.org/licenses/BSD-3-Clause

from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum, IntEnum, auto
from typing import ClassVar, List, Optional, Sequence

import numpy as np
import numpy.typing as npt
from astropy.coordinates import Angle
from astropy.units import Quantity

from lucupy.helpers import flatten
from lucupy.types import ScalarOrNDArray

from ..decorators import immutable
from .timingwindow import TimingWindow


class SkyBackground(float, Enum):
    """
    Bins for observation sky background requirements or current conditions.

    Members:
        - SB20 = 0.2
        - SB50 = 0.5
        - SB80 = 0.8
        - SBANY = 1.0

    """
    SB20 = 0.2
    SB50 = 0.5
    SB80 = 0.8
    SBANY = 1.0


class CloudCover(float, Enum):
    """
    Bins for observation cloud cover requirements or current conditions.

    Members:
        - CC50 = 0.5
        - CC70 = 0.7
        - CC80 = 0.8
        - CCANY = 1.0
    """
    CC50 = 0.5
    CC70 = 0.7
    CC80 = 0.8
    CCANY = 1.0


class ImageQuality(float, Enum):
    """
    Bins for observation image quality requirements or current conditions.

    Members:
        - IQ20 = 0.2
        - IQ70 = 0.7
        - IQ85 = 0.85
        - IQANY = 1.0

    """
    IQ20 = 0.2
    IQ70 = 0.7
    IQ85 = 0.85
    IQANY = 1.0


class WaterVapor(float, Enum):
    """
    Bins for observation water vapor requirements or current conditions.

    Members:
        - WV20 = 0.2
        - WV50 = 0.5
        - WV80 = 0.8
        - WVANY = 1.0

    """
    WV20 = 0.2
    WV50 = 0.5
    WV80 = 0.8
    WVANY = 1.0


class Strehl(float, Enum):
    """
    The Strehl ratio is a measure of the quality of optical image formation.
    Used variously in situations where optical resolution is compromised due to lens aberrations or due to imaging
    through the turbulent atmosphere, the Strehl ratio has a value between 0 and 1, with a hypothetical, perfectly
    unaberrated optical system having a Strehl ratio of 1. (Source: Wikipedia.)

    Members:
        - S00 = 0.0
        - S02 = 0.2
        - S04 = 0.4
        - S06 = 0.6
        - S08 = 0.8
        - S10 = 1.0

    """
    S00 = 0.0
    S02 = 0.2
    S04 = 0.4
    S06 = 0.6
    S08 = 0.8
    S10 = 1.0


class ElevationType(IntEnum):
    """
    The type of elevation constraints in the observing conditions.

    Members:
       - NONE
       - HOUR_ANGLE
       - AIRMASS
    """
    NONE = auto()
    HOUR_ANGLE = auto()
    AIRMASS = auto()


@immutable
@dataclass(frozen=True)
class Conditions:
    """
    A set of conditions.

    Note that we make this dataclass eq and ordered so that we can compare one
    set of conditions with another to see if one satisfies the other.

    This should be done via:
    current_conditions <= required_conditions.

    Attributes:
        cc (ScalarOrNDArray[CloudCover]): CloudCover
        iq (ScalarOrNDArray[ImageQuality]): ImageQuality
        sb (ScalarOrNDArray[SkyBackground]): SkyBackground
        wv (ScalarOrNDArray[WaterVapor]): WaterVapor

    """
    cc: ScalarOrNDArray[CloudCover]
    iq: ScalarOrNDArray[ImageQuality]
    sb: ScalarOrNDArray[SkyBackground]
    wv: ScalarOrNDArray[WaterVapor]

    # Least restrictive conditions.
    @classmethod
    def least_restrictive(cls) -> Conditions:
        """
        Return the least possible restrictive conditions.
        """
        return cls(cc=CloudCover.CCANY,
                   iq=ImageQuality.IQANY,
                   sb=SkyBackground.SBANY,
                   wv=WaterVapor.WVANY)

    def __post_init__(self):
        """
        Ensure that if any arrays are specified, all values are specified arrays of the same size.
        """
        is_uniform = len({np.isscalar(self.cc), np.isscalar(self.iq), np.isscalar(self.sb), np.isscalar(self.wv)}) == 1
        if not is_uniform:
            raise ValueError(f'Conditions have a mixture of array and scalar types: {self}')

        are_arrays = isinstance(self.cc, np.ndarray)
        if are_arrays:
            uniform_lengths = len({self.cc.size, self.iq.size, self.sb.size, self.wv.size}) == 1
            if not uniform_lengths:
                raise ValueError(f'Conditions have a variable number of array sizes: {self}')

    @staticmethod
    def most_restrictive_conditions(conditions: Sequence[Conditions]) -> Conditions:
        """
        Given an iterable of conditions, find the most restrictive amongst the set.
        If no conditions are given, return the most flexible conditions possible.
        """
        if len(conditions) == 0:
            return Conditions.least_restrictive()
        min_cc = min(flatten(c.cc for c in conditions), default=CloudCover.CCANY)
        min_iq = min(flatten(c.iq for c in conditions), default=ImageQuality.IQANY)
        min_sb = min(flatten(c.sb for c in conditions), default=SkyBackground.SBANY)
        min_wv = min(flatten(c.wv for c in conditions), default=WaterVapor.WVANY)
        return Conditions(cc=min_cc, iq=min_iq, sb=min_sb, wv=min_wv)

    def __len__(self):
        """
        For array values, return the length of the arrays.
        For scalar values, return a length of 1.
        """
        return len(self.cc) if isinstance(self.cc, np.ndarray) else 1


@immutable
@dataclass(frozen=True)
class Constraints:
    """The constraints required for an observation to be performed.

    Default airmass values to use for elevation constraints if:
        1. The Constraints are not present in the Observation at all; or
        2. The elevation_type is set to NONE.

    Attributes:
        conditions (Conditions): Collection of conditions.
        elevation_type (ElevationType): Elevation type.
        elevation_min (float): Max value of elevation.
        elevation_max (float): Min value of elevation.
        timing_windows (List[TimingWindow]): Time windows in the constraints are in effect.
        strehl (Strehl:optional): None

        DEFAULT_AIRMASS_ELEVATION_MIN (ClassVar[float]): 1.0
        DEFAULT_AIRMASS_ELEVATION_MAX (ClassVar[float]): 2.3

    """
    conditions: Conditions
    # constrast: Constrast
    elevation_type: ElevationType
    elevation_min: float
    elevation_max: float
    timing_windows: List[TimingWindow]
    # clearance_windows: Optional[List[ClearanceWindow]] = None
    strehl: Optional[Strehl] = None

    # Default airmass values to use for elevation constraints if:
    # 1. The Constraints are not present in the Observation at all; or
    # 2. The elevation_type is set to NONE.
    DEFAULT_AIRMASS_ELEVATION_MIN: ClassVar[float] = field(init=False, default=1.0, repr=False, compare=False)
    DEFAULT_AIRMASS_ELEVATION_MAX: ClassVar[float] = field(init=False, default=2.3, repr=False, compare=False)


@immutable
@dataclass(frozen=True)
class Variant:
    """A weather variant.
    wind_speed should be in m / s.

    Attributes:
        iq (Union[npt.NDArray[ImageQuality], ImageQuality]): Image quality.
        cc (Union[npt.NDArray[CloudCover], CloudCover]): Cloud Cover.
        wind_dir (Angle): Wind direction.
        wind_spd (Quantity): Wind speed.

    """
    # TODO: No idea what time blocks are. Note this could be a list or a single value.
    # TODO: Because of this, we cannot hash Variants, which is problematic.

    iq: npt.NDArray[ImageQuality] | ImageQuality
    cc: npt.NDArray[CloudCover] | CloudCover
    wind_dir: Angle
    wind_spd: Quantity

    def __post_init__(self):
        """
        Ensure that if any arrays are specified, all values are specified arrays of the same size.
        """
        is_uniform = len({np.isscalar(self.cc), np.isscalar(self.iq)}) == 1
        if not is_uniform:
            raise ValueError(f'Variant has a mixture of array and scalar types: {self}')

        are_arrays = isinstance(self.cc, np.ndarray)
        array_lengths = {np.asarray(self.wind_dir).size,
                         np.asarray(self.wind_spd).size}
        if are_arrays:
            uniform_lengths = len({len(self.cc), len(self.iq)}.union(array_lengths)) == 1
        else:
            uniform_lengths = len(array_lengths) == 1
        if not uniform_lengths:
            raise ValueError(f'Variant has a variable number of array sizes: {self}')



# Copyright (c) 2016-2023 Association of Universities for Research in Astronomy, Inc. (AURA)
# For license information see LICENSE or https://opensource.org/licenses/BSD-3-Clause

from __future__ import annotations

from dataclasses import dataclass
from datetime import timedelta
from enum import IntEnum, auto
from typing import List, Mapping, Optional

import numpy as np

from lucupy.observatory.abstract import ObservatoryProperties

from ..types import ZeroTime
from .atom import Atom
from .constraints import Constraints
from .ids import ObservationID, ProgramID, UniqueGroupID
from .qastate import QAState
from .resource import NIR_INSTRUMENTS, Resource, Resources
from .site import Site
from .target import Target, TargetType
from .too import TooType
from .wavelength import Wavelengths


class ObservationStatus(IntEnum):
    """
    The status of an observation as indicated in the Observing Tool / ODB.

    Members:
        - NEW
        - INCLUDED
        - PROPOSED
        - APPROVED
        - FOR_REVIEW: Not in original mini-model description, but returned by OCS.
        - ON_HOLD
        - READY
        - ONGOING
        - OBSERVED
        - INACTIVE
        - PHASE2
    """
    NEW = auto()
    INCLUDED = auto()
    PROPOSED = auto()
    APPROVED = auto()
    FOR_REVIEW = auto()
    ON_HOLD = auto()
    READY = auto()
    ONGOING = auto()
    OBSERVED = auto()
    INACTIVE = auto()
    PHASE2 = auto()


class Priority(IntEnum):
    """An observation's priority.
    Note that these are ordered specifically so that we can compare them.

    Members:
        - LOW
        - MEDIUM
        - HIGH
    """
    LOW = auto()
    MEDIUM = auto()
    HIGH = auto()


class SetupTimeType(IntEnum):
    """The setup time type for an observation.

    Members:
        - NONE
        - REACQUISITION
        - FULL
    """
    NONE = auto()
    REACQUISITION = auto()
    FULL = auto()


class ObservationClass(IntEnum):
    """The class of an observation.

    Note that the order of these is specific and deliberate: they are listed in
    preference order for observation classes, and hence, should not be rearranged.
    These correspond to the values in the OCS when made uppercase.

    Members:
        - SCIENCE
        - PROGCAL
        - PARTNERCAL
        - ACQ
        - ACQCAL
        - DAYCAL

    """
    SCIENCE = auto()
    PROGCAL = auto()
    PARTNERCAL = auto()
    ACQ = auto()
    ACQCAL = auto()
    DAYCAL = auto()
    NONE = auto()


@dataclass
class Observation:
    """Representation of an observation.
    Non-obvious fields are documented below.
    Attributes:
        id (ObservationID): should represent the observation's ID, e.g. GN-2018B-Q-101-123.
        internal_id (str): is the key associated with the observation
        order (int): refers to the order of the observation in either its group or the program
        title (str):
        site (site): the Site in which the Observation has to be done.
        status (ObservationStatus):
        active (bool): Indicates if the Observation is active or not.
        priority(Priority): Priority of the Observation. Affects the scoring.
        setuptime_type(SetupTimeType): When doing / resuming an Observation,
            indicates if a reacquisition, a full setup or just nothing has to be done.
        acq_overhead (timedelta): Time overhead for acquisition.
        obs_class (ObservationClass): Type of Observation.
        targets (List[Target]): should contain a complete list of all targets associated with the observation,
            with the base being in the first position
        guiding (Mapping[Resource, Target]): is a map between guide probe resources and their targets.
        sequence (List[Atom]): Sequence of Atoms that describe the observation.
        belongs_to (ProgramID): ID for the program the observation belongs to.
        constraints (Constraints, optional): Some observations do not have constraints, e.g. GN-208A-FT-103-6.
        too_type (TooType, optional): Default to None.

    """
    id: ObservationID
    internal_id: str
    order: int
    title: str
    site: Site
    status: ObservationStatus
    active: bool
    priority: Priority
    setuptime_type: SetupTimeType
    acq_overhead: timedelta

    # TODO: This will be handled differently between OCS and GPP.
    # TODO: 1. In OCS, when the sequence is examined, the ObservationClasses of the
    # TODO:    individual observes (sequence steps) will be analyzed and the highest
    # TODO:    precedence one will be set for the observation (based on the earliest
    # TODO:    ObservationClass in the enum).
    # TODO: 2. In GPP, this information will be handled automatically and require no
    # TODO:    special processing.
    # TODO: Should this be Optional?
    obs_class: ObservationClass

    targets: List[Target]
    guiding: Mapping[Resource, Target]
    sequence: List[Atom]
    belongs_to: ProgramID

    # Some observations do not have constraints, e.g. GN-208A-FT-103-6.
    # to mypy complaince we should have an EmptyConstraints
    constraints: Constraints

    too_type: Optional[TooType] = None

    @property
    def to_unique_group_id(self) -> UniqueGroupID:
        """
        A method to return the UniqueGroupID that contains this observation.
        """
        return self.id.to_unique_group_id

    @property
    def unique_id(self) -> ObservationID:
        """
        Unique ID for the Observation to amalgamate all the IDs from Group down to Observation.
        This way, for any group, group.children will return an ID, giving a mixed list of UniqueGroupID and
        ObservationID, which will be used to select the top level groups.
        """
        return self.id

    def base_target(self) -> Optional[Target]:
        """
        Returns:
            Get the base target for this Observation if it has one, and None otherwise.
        """
        def filter_by_type(t: Optional[Target]):
            if t is not None:
                return t.type == TargetType.BASE

        return next(filter(lambda t: filter_by_type(t), self.targets), None)

    def exec_time(self) -> timedelta:
        """
        Returns:
            Total execution time for the program, which is the sum across atoms and the acquisition overhead.
        """
        return sum((atom.exec_time for atom in self.sequence), ZeroTime) + self.acq_overhead

    def total_used(self) -> timedelta:
        """
        Returns:
            Total program time used: includes program time and partner time.
        """
        return self.program_used() + self.partner_used()

    def required_resources(self) -> Resources:
        """
        Returns:
            The required resources for an observation based on the sequence's needs.
        """
        # TODO: For now, we do not return guiding keys amongst the resources.
        # return frozenset(self.guiding.keys() | {r for a in self.sequence for r in a.resources})
        return frozenset((r for a in self.sequence for r in a.resources))

    def instrument(self) -> Optional[Resource]:
        """
        Returns:
            A resource that is an instrument, if one exists. There should be only one.
        """
        def check_instrument(r: Optional[Resource]):
            if r is not None:
                return ObservatoryProperties.is_instrument(r)

        return next(filter(lambda r: check_instrument(r),
                           self.required_resources()), None)

    def is_nir(self) -> bool:
        """Define if the observation is a NIR observation or not."""
        inst_req = self.instrument()
        return any(inst_req == nir_ins for nir_ins in NIR_INSTRUMENTS)

    def wavelengths(self) -> Wavelengths:
        """
        Returns:
            The set of wavelengths included in the sequence.
        """
        return frozenset(w for c in self.sequence for w in c.wavelengths)

    def program_used(self) -> timedelta:
        """We roll this information up from the atoms as it will be calculated
            during the Optimizer algorithm. Note that it is also available directly
            from the OCS, which is used to populate the time allocation.
        Returns:
            (timedelta): With the time program used.
        """
        return sum((atom.program_used for atom in self.sequence), start=ZeroTime)

    def partner_used(self) -> timedelta:
        """We roll this information up from the atoms as it will be calculated
        during the Optimizer algorithm. Note that it is also available directly
        from the OCS, which is used to populate the time allocation.

        Returns:
            (timedelta): With the time partner used.
        """
        return sum((atom.partner_used for atom in self.sequence), start=ZeroTime)

    def prog_time(self) -> timedelta:
        """We roll this information up from the atoms.

        Returns:
            (timedelta): With the total planned program time.
        """
        return sum((atom.prog_time for atom in self.sequence), start=ZeroTime)

    def part_time(self) -> timedelta:
        """We roll this information up from the atoms.

        Returns:
            (timedelta): With the total planned partner time.
        """
        return sum((atom.part_time for atom in self.sequence), start=ZeroTime)

    def cumulative_exec_times(self) -> List[timedelta]:
        """Cumulative series of execution times for the unobserved atoms
        in a sequence, excluding acquisition time."""
        cum_seq = [atom.exec_time if not atom.observed else timedelta() for atom in self.sequence]
        return np.cumsum(cum_seq)

    @staticmethod
    def _select_obsclass(classes: List[ObservationClass]) -> Optional[ObservationClass]:
        """Given a list of non-empty ObservationClasses, determine which occurs with
        the highest precedence in the ObservationClasses enum, i.e. has the lowest index.

        This will be used when examining the sequence for atoms.

        Returns:
            (ObservationClass): Lowest-index ObservationClasses or None if the list is empty.
        """
        # TODO: Move this to the ODB program extractor as the logic is used there.
        # TODO: Remove from Bryan's atomizer.
        return min(classes, default=None)

    @staticmethod
    def _select_qastate(qastates: List[QAState]) -> Optional[QAState]:
        """
        Given a list of non-empty QAStates, determine which occurs with the
        highest precedence in the QAStates enum, i.e. has the lowest index.

        Returns:
            (QAState): Lowest-index QAStates or None if the list is empty.
        """
        # TODO: Move this to the ODB program extractor as the logic is used there.
        # TODO: Remove from Bryan's atomizer.

        return min(qastates, default=None)

    def show(self, depth: int = 1) -> None:
        """Print content of the Observation.

        Args:
            depth (int, optional): depth of the separator. Defaults to 1.
        """
        def sep(indent: int) -> str:
            return '-----' * indent

        print(f'{sep(depth)} Observation: {self.id.id}')
        for atom in self.sequence:
            print(f'{sep(depth + 1)} {atom}')

    def __len__(self):
        """
        This is to treat observations the same as groups and is a bit of a hack.
        Observations are to be placed in AND Groups of size 1 for scheduling purposes.
        """
        return 1

    def __eq__(self, other: Observation) -> bool:  # type: ignore[override]
        """
        We override the equality checker created by @dataclass to temporarily skip sequence
        comparison in test cases until the atom creation process is finish.
        """

        return (dict((k, v) for k, v in self.__dict__.items() if k != 'sequence') ==
                dict((k, v) for k, v in other.__dict__.items() if k != 'sequence'))



# All code in this package is a refactored, numpy-vectorized version of thorskyutil.py:
#
# https://github.com/jrthorstensen/thorsky/blob/master/thorskyutil.py
#
# utility and miscellaneous time and the sky routines built mostly on astropy.
#
# Copyright John Thorstensen, 2018, who graciously has allowed Gemini to use this code under the BSD-3 Clause license.
# For license information see LICENSE or https://opensource.org/licenses/BSD-3-Clause

from typing import Tuple, TypeVar, Union

import astropy.units as u
import numpy as np
import numpy.typing as npt
from astropy.coordinates import Angle, Longitude

T = TypeVar('T')
U = TypeVar('U')
ScalarOrArray = Union[T, U, npt.NDArray[U]]
AngleParam = ScalarOrArray[Angle, float]


class Altitude:
    def __init__(self):
        raise NotImplementedError('Use static method Altitude.above.')

    @staticmethod
    def above(dec: AngleParam,
              ha: AngleParam,
              lat: AngleParam) -> Tuple[Angle, Angle, Angle]:
        """Compute altitude above horizon, azimuth, and parallactic angle.

        The parallactic angle is the position angle of the arc between the
        object and the zenith, i.e. the position angle that points 'straight up'
        when you're looking at an object.  It is important for atmospheric
        refraction and dispersion compensation, which Filippenko discusses
        ( 1982PASP...94..715F ).  Does not take small effects into
        account (polar motion, nutation, whatever) so is lighter weight than
        the astropy equivalents.

        Filippenko's expression for the parallactic angle leaves it to the user
        to select the correct root of an inverse trig function; this is handled
        automatically here by fully solving the astronomical triangle.

        The astropy altaz transformation depends on the 3 Mbyte download from USNO
        to find the lst, so here is a stripped down version.
        # Arguments are all assumed to be Angles so they don't need to be converted to radians;
        the dec is assumed to be in equinox of date to avoid
        # We get the parallactic angle almost for since we have ha, colat, and altitude.

        Args:
            dec: Declination
            ha: Hour angle (spherical astronomy) of the position, positive westward
            lat: Latitude of site.

        Returns:
            tuple of (altitude, azimuth, parallactic), all of which are Angles.

        """

        dec = np.asarray(dec.to_value(u.rad).data) * u.rad  # type: ignore
        ha = np.asarray(ha.to_value(u.rad)) * u.rad  # type: ignore
        scalar_input = False
        if dec.ndim == 0 and ha.ndim == 0:
            scalar_input = True
        if dec.ndim == 0:
            dec = dec[None]
        if ha.ndim == 0:
            ha = ha[None]

        if len(dec) == 1 and len(ha) > 1:
            dec = dec * np.ones(len(ha))
        elif len(dec) > 1 and len(ha) == 1:
            ha = ha * np.ones(len(dec))
        elif len(dec) != len(ha):
            raise ValueError('Error: dec and ha have incompatible lengths')

        cos_dec = np.cos(dec)
        sin_dec = np.sin(dec)
        cos_ha = np.cos(ha)
        sin_ha = np.sin(ha)
        cos_lat = np.cos(lat)
        sinlat = np.sin(lat)

        # Compute the altitude
        altitude = Angle(np.arcsin(cos_dec * cos_ha * cos_lat + sin_dec * sinlat), unit=u.radian)

        # Compute the azimuth
        y = sin_dec * cos_lat - cos_dec * cos_ha * sinlat  # due north component
        z = -1. * cos_dec * sin_ha  # due east component
        azimuth = Longitude(np.arctan2(z, y), unit=u.radian)

        # Now solve the spherical trig to give parallactic angle
        parang = np.zeros(len(dec))
        ii = np.where(cos_dec != 0.0)[0][:]
        if len(ii) != 0:
            # spherical law of sines
            sinp = -1. * np.sin(azimuth[ii]) * cos_lat / cos_dec[ii]
            # spherical law of cosines, also expressed in already computed variables.
            cosp = -1. * np.cos(azimuth[ii]) * cos_ha[ii] - np.sin(azimuth[ii]) * sin_ha[ii] * sinlat
            parang[ii] = np.arctan2(sinp, cosp)

        jj = np.where(cos_dec == 0.0)[0][:]  # you're on the pole
        if len(jj) != 0:
            parang[jj] = np.pi

        if scalar_input:
            altitude = np.squeeze(altitude)
            azimuth = np.squeeze(azimuth)
            parang = np.squeeze(parang)
        return altitude, azimuth, Angle(parang, unit=u.rad)



# Copyright (c) 2016-2022 Association of Universities for Research in Astronomy, Inc. (AURA)
# For license information see LICENSE or https://opensource.org/licenses/BSD-3-Clause

from __future__ import annotations

from dataclasses import dataclass
from datetime import date
from enum import Enum
from typing import Dict, List

from gelidum import freeze

from ..decorators import immutable


class SemesterHalf(str, Enum):
    """Gemini typically schedules programs for two semesters per year, namely A and B.
    For other observatories, this logic might have to be substantially changed.

    Members:
        - A = 'A'
        - B = 'B'
    """
    A = 'A'
    B = 'B'

    def start_day(self) -> int: # noqa
        return 1

    def end_day(self) -> int: # noqa
        return 31

    def start_month(self) -> int:
        return _semester_half_months[self][0]

    def end_month(self) -> int:
        return _semester_half_months[self][-1]

    @staticmethod
    def determine_half(month: int) -> SemesterHalf:
        """
        Given a month, return the SemesterHalf this falls in.

        Args:
            month: the month of the date

        Returns:
            The SemesterHalf in which this month occurs.
        """
        if month < 1 or month > 12:
            raise ValueError(f'Illegal month: {month}')
        for semester_half, months in _semester_half_months.items():
            if month in months:
                return semester_half
        raise ValueError(f'Month {month} cannot be found in any SemesterHalf.')


# Make a FrozenDict with values FrozenList, so the lists cannot be changed, nor can the dict.
_semester_half_months: Dict[SemesterHalf, List[int]] = freeze({
    SemesterHalf.A: [2, 3, 4, 5, 6, 7],
    SemesterHalf.B: [8, 9, 10, 11, 12, 1]
})


@immutable
@dataclass(frozen=True, order=True)
class Semester:
    """ A semester is a period for which programs may be submitted to Gemini.

    Attributes:
        year (int): A four digit year.
        half (SemesterHalf): Two semesters during each year, indicated by the SemesterHalf.
    """
    year: int
    half: SemesterHalf

    def start_date(self) -> date:
        """
        Returns:
            The start date of the Semester.
        """
        return date(year=self.year, month=self.half.start_month(), day=self.half.start_day())

    def end_date(self) -> date:
        """
        Returns:
            The end date of the Semester.
        """
        # Check if the semester spans two years.
        if _semester_half_months[self.half][-1] < _semester_half_months[self.half][0]:
            end_year = self.year + 1
        else:
            end_year = self.year
        return date(year=end_year, month=self.half.end_month(), day=self.half.end_day())

    @staticmethod
    def find_semester_from_date(lookup_date: date) -> Semester:
        """
        Given a date, return the Semester in which it occurs.

        Args:
            A date object representing the date.

        Returns:
            The Semester in which this date occurs.
        """
        # Determine which semester half the date falls in.
        semester_half = SemesterHalf.determine_half(lookup_date.month)

        # If the month in the semester half is less than the first entry, it spans over a year.
        if lookup_date.month < _semester_half_months[semester_half][0]:
            semester_year = lookup_date.year - 1
        else:
            semester_year = lookup_date.year

        return Semester(year=semester_year, half=semester_half)

    def __str__(self):
        return f'{self.year}{self.half.value}'
